import logging
import sys
import numpy as np
from scipy.sparse import csr_matrix
from patsy import dmatrix
import sympy



def _design_matrix(formula, data_df):
    """
    Build the design matrix
    :param formula:  An object that can be used to construct a design matrix. See patsy.dmatrix for detail.
    :param data_df: Pandas data frame
    :return: full rank design matrix
    """
    dmat = dmatrix(formula, data_df)
    dmat = np.asarray(dmat)
    return dmat


def _design_matrix_crank(formula, data_df):
    """
    Build the design matrix
    :param formula:  An object that can be used to construct a design matrix. See patsy.dmatrix for detail.
    :param data_df: Pandas data frame
    :return: design matrix, full rank design matrix, the indexes for the linear independent columns, label of design matrix
    """
    dmat = dmatrix(formula, data_df)
    col_arr = repr(dmat).split("\n")[1].split('\s+')
    dmat = np.asarray(dmat)
    _, indexes = sympy.Matrix(dmat).rref()
    dmat_full_rank = dmat[:, indexes]
    return dmat, dmat_full_rank, indexes, col_arr



import numpy as np
import pandas as pd
import logging
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy.stats import beta


def qqplot(gwas_res_file, p_colname, CI=True, out_file='QQ.png'):
    """
    Q-Q plot from the GWAS results
    :param gwas_res_file: the result file of GWAS
    :param p_colname: the column name of p values
    :param CI: Whether to draw the confidence interval. Default is True.
    :param out_file: the output file. Default is QQ.png
    :return: 0
    """
    df = pd.read_csv(gwas_res_file, header=0, sep='\s+')
    p_obs = np.sort(df[p_colname])
    p_obs = p_obs[~np.isnan(p_obs)]
    arr = np.arange(1, len(p_obs) + 1)
    p_exp = (arr-0.5)/len(arr)
    logging.info("The median of p values is: {}".format(np.median(p_obs)))
    plt.figure(figsize=(3.5, 3.5), dpi=600)
    plt.rc('font', family='Times New Roman')
    plt.xlabel('Expected -log10($P$)', fontsize=10)
    plt.ylabel('Observed -log10($P$)', fontsize=10)
    plt.xticks(fontsize=8)
    plt.yticks(fontsize=8)
    if CI:
        p_up = beta.ppf(0.025, arr, arr[::-1])
        p_low = beta.ppf(0.975, arr, arr[::-1])
        plt.fill_between(-np.log10(p_exp), -np.log10(p_up), -np.log10(p_low), color='grey')
    plt.scatter(-np.log10(p_exp), -np.log10(p_obs), s=1.5, c='red')
    plt.plot(-np.log10(p_exp), -np.log10(p_exp), 'k')
    plt.savefig(out_file)
    return 0



import logging
import numpy as np
import pandas as pd
from scipy import linalg
from pysnptools.snpreader import Bed
from tqdm import tqdm
from gmat.process_plink.process_plink import impute_geno
from scipy.stats import chi2
from scipy.sparse import csc_matrix
from patsy import dmatrix


def fixed_(pheno_file, bed_file, agmat_file, class_lst=None, fix=None, out_file='res.txt', condition_snp=None, npart=100,
          maxiter0=100, maxiter1=10, speed=True, pcut=5.0e-5):
    """
    run GWAS based on linear mixed model and put one snp in each analysis.
    :param pheno_file: phenotypic file with head rows. The fist column is individual id. The last column is phenotypic values.
    Class variates and covariates are in the middle columns. Missing values in the phenotypic file can be
    expressed as 'NA', 'NaN', 'nan', 'na'.
    :param bed_file: the prefix for the plink binary file.
    :param agmat_file: the prefix for the additive genomic relationship matrix file.
    :param class_lst: a list for class variates
    :param fix: formula for the fix effect. Eg. fix='sex + herd'
    :param out_file: the output file. Default is 'res.txt'.
    :param condition_snp: A string or a list for the conditional SNP. Default is None.
    :param npart: divide the SNPs into n parts and read them into memory successively. Default is 100.
    :param maxiter0: the maximum number of iteration for variance estimation in the null model
    :param maxiter1: the maximum number of iteration for variance estimation in testing snp.
    :param speed: whether to speed up with estimated variances in the null model.
    :param pcut: cut p values to select top snps to re-estimate the p values.
    :return:
    """
    logging.info("###Prepare the phenotypic values, covariates and genomic matrix")
    y, xmat, zmat, ag, id_keep = _pre_pheno(pheno_file, agmat_file, class_lst, fix)
    logging.info("###Estimate the variances in the null model.")
    init = np.array([np.var(y) / 2] * 2)
    gmat_eigenvals, gmat_eigenvecs = linalg.eigh(zmat.dot((zmat.dot(ag)).T))
    gmat_eigenvals = np.array(gmat_eigenvals).reshape(-1, 1)
    y_trans = np.dot(gmat_eigenvecs.T, y)  # transform the phenotypic values
    xmat_trans = np.dot(gmat_eigenvecs.T, xmat)  # transform the fixed effects
    var, convergence = _weight_emai_eigen(y_trans, xmat_trans, gmat_eigenvals, init=init, maxiter=maxiter0, cc=1.0e-8)
    logging.info('Whether converge?: {}'.format(convergence))
    logging.info('The estimated variances: {}'.format(var))
    snp_on_disk = Bed(bed_file, count_A1=False)
    num_snp = snp_on_disk.sid_count
    fam_df = pd.read_csv(bed_file + '.fam', header=None, sep='\s+')
    id_in_fam = list(np.array(fam_df.iloc[:, 1], dtype=np.str))
    id_keep_ind = []
    for val in id_keep:
        id_keep_ind.append(id_in_fam.index(val))
    snp_index = []
    for i in range(npart):
        snp_index.append(int(num_snp / npart) * i)
    snp_index.append(num_snp)
    logging.info("Conditional SNP: {}".format(condition_snp))
    if condition_snp is not None:
        if not isinstance(condition_snp, list):
            condition_snp = [condition_snp]
        condition_snp_index = snp_on_disk.sid_to_index(condition_snp)
        condition_snp_val = snp_on_disk[id_keep_ind, condition_snp_index].read().val
        if np.any(np.isnan(condition_snp_val)):
            logging.info('Missing genotypes of conditional SNP are imputed with random genotypes.')
            condition_snp_val = impute_geno(condition_snp_val)
        condition_snp_val = np.dot(gmat_eigenvecs.T, condition_snp_val)
        xmat_trans = np.concatenate([xmat_trans, condition_snp_val], axis=1)
    logging.info('###Start the association')
    with open(out_file, 'w') as fout:
        fout.write('\t'.join(['chro', 'snp_ID', 'pos', 'allele1', 'allele2', 'cc', 'eff', 'se', 'p_wald']))
        fout.write('\n')
    maxiter = maxiter1
    if not convergence:
        maxiter = 0
    if speed:  # whether to speed up with estimated variances in the null model
        vmat = 1.0 / (gmat_eigenvals * var[0] + var[1])
        vx = np.multiply(vmat, xmat_trans)
        xvx = np.dot(xmat_trans.T, vx)
        xvx = linalg.inv(xvx)
        # py
        pmat = -np.dot(np.dot(vx, xvx), vx.T)
        diag_val = np.diag(pmat)
        np.fill_diagonal(pmat, diag_val + vmat[:, -1])
        py = np.dot(pmat, y_trans)
        for i in range(npart):
            logging.info('The {}/{} part'.format(i + 1, npart))
            snp_info = pd.read_csv(bed_file + '.bim', sep='\s+', header=None, skiprows=snp_index[i],
                                   nrows=snp_index[i + 1] - snp_index[i])
            res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
            res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
            snp_mat = snp_on_disk[id_keep_ind, snp_index[i]:snp_index[i + 1]].read().val
            if np.any(np.isnan(snp_mat)):
                snp_mat = impute_geno(snp_mat)
            snp_mat = np.dot(gmat_eigenvecs.T, snp_mat)
            se_arr = np.sum(np.multiply(snp_mat, np.dot(pmat, snp_mat)), axis=0)
            se_arr = 1/se_arr
            eff_arr = np.dot(snp_mat.T, py)[:, -1]
            p_arr = chi2.sf(eff_arr * eff_arr * se_arr, 1)
            eff_arr = eff_arr * se_arr
            se_arr = np.sqrt(se_arr)
            cc_arr = np.array([1000.0]*len(se_arr))
            ind_bool = p_arr < pcut
            cc_vec = []
            eff_vec = []
            se_vec = []
            p_vec = []
            snp_mat = snp_mat[:, ind_bool]
            for k in tqdm(range(snp_mat.shape[1])):
                snpk = snp_mat[:, k:(k + 1)]
                xmat_trans_k = np.concatenate([xmat_trans, snpk], axis=1)
                cc_val, snp_eff, se, p_val = _weight_emai_eigen_snp(y_trans, xmat_trans_k, gmat_eigenvals, var,
                                                                    maxiter=maxiter, cc=1.0e-8)
                cc_vec.append(cc_val)
                eff_vec.append(snp_eff)
                se_vec.append(se)
                p_vec.append(p_val)
            cc_arr[ind_bool] = cc_vec
            eff_arr[ind_bool] = eff_vec
            se_arr[ind_bool] = se_vec
            p_arr[ind_bool] = p_vec
            res_df.loc[:, 'cc_val'] = cc_arr
            res_df.loc[:, 'eff_val'] = eff_arr
            res_df.loc[:, 'se_val'] = se_arr
            res_df.loc[:, 'p_val'] = p_arr
            res_df.to_csv(out_file, sep='\t', index=False, header=False, mode='a', na_rep='NA')
    else:
        for i in range(npart):
            logging.info('The {}/{} part'.format(i + 1, npart))
            snp_info = pd.read_csv(bed_file + '.bim', sep='\s+', header=None, skiprows=snp_index[i],
                                   nrows=snp_index[i + 1] - snp_index[i])
            res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
            res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
            snp_mat = snp_on_disk[id_keep_ind, snp_index[i]:snp_index[i + 1]].read().val
            if np.any(np.isnan(snp_mat)):
                snp_mat = impute_geno(snp_mat)
            snp_mat = np.dot(gmat_eigenvecs.T, snp_mat)
            cc_vec = []
            eff_vec = []
            se_vec = []
            p_vec = []
            for k in tqdm(range(snp_mat.shape[1])):
                snpk = snp_mat[:, k:(k + 1)]
                xmat_trans_k = np.concatenate([xmat_trans, snpk], axis=1)
                cc_val, snp_eff, se, p_val = _weight_emai_eigen_snp(y_trans, xmat_trans_k, gmat_eigenvals, var,
                                                                    maxiter=maxiter, cc=1.0e-8)
                cc_vec.append(cc_val)
                eff_vec.append(snp_eff)
                se_vec.append(se)
                p_vec.append(p_val)
            res_df.loc[:, 'cc_val'] = cc_vec
            res_df.loc[:, 'eff_val'] = eff_vec
            res_df.loc[:, 'se_val'] = se_vec
            res_df.loc[:, 'p_val'] = p_vec
            res_df.to_csv(out_file, sep='\t', index=False, header=False, mode='a', na_rep='NA')
    return 0


def _pre_pheno(pheno_file, agmat_file, class_lst, fix):
    id_in_agmat = []
    with open(agmat_file + '.id') as fin:
        for line in fin:
            arr = line.split()
            id_in_agmat.append(arr[0])
    data_df = pd.read_csv(pheno_file, sep='\s+', header=0)
    data_df = data_df.dropna()
    col_names = data_df.columns
    class_lst2 = [col_names[0]]
    if class_lst is not None:
        class_lst2 = class_lst + class_lst2
    for val in col_names:
        if val in class_lst2:
            data_df[val] = data_df[val].astype('str')
        else:
            data_df[val] = data_df[val].astype('float')
    id_keep = []  # keep the id both in the agmat file and pheno file, without missing phenotypes and covariates.
    for i in range(data_df.shape[0]):
        if data_df.iloc[i, 0] in id_in_agmat:
            id_keep.append(True)
        else:
            id_keep.append(False)
    data_df = data_df.iloc[id_keep, :]
    id_keep = list(data_df.iloc[:, 0])
    y = np.array(data_df.iloc[:, -1]).reshape(-1, 1)
    xmat = np.ones((y.shape[0], 1))
    if fix is not None:
        xmat = dmatrix(fix, data_df)
        xmat = np.asarray(xmat)
        _, r = np.linalg.qr(xmat)
        indexes = np.absolute(np.diag(r)) >= 1e-10
        xmat = xmat[:, indexes]
    val = 0
    id_dct = {}
    col = []
    for v in id_keep:
        if v not in id_dct:
            id_dct[v] = val
            val += 1
        col.append(id_dct[v])
    row = range(len(id_keep))
    zmat = csc_matrix(([1.0]*len(row), (row, col)))
    ag = np.zeros((len(id_dct), len(id_dct)))
    with open(agmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    return y, xmat, zmat, ag, id_keep


def _weight_emai_eigen(y_trans, xmat_trans, gmat_eigenvals, init=None, maxiter=100, cc=1.0e-8):
    fd_mat = np.zeros(2)
    ai_mat = np.zeros((2, 2))
    em_mat = np.zeros((2, 2))
    num_id = gmat_eigenvals.shape[0]
    cc_val = 1000.0
    var = np.array(init)
    var_update = var * 1000
    delta = var_update - var
    convergence = False
    for i in range(maxiter):
        logging.info("Iteration: {}".format(i+1))
        vmat = 1.0 / (gmat_eigenvals * var[0] + var[1])
        vx = np.multiply(vmat, xmat_trans)
        xvx = np.dot(xmat_trans.T, vx)
        xvx = linalg.inv(xvx)
        # py
        xvy = np.dot(vx.T, y_trans)
        y_xb = y_trans - np.dot(xmat_trans, np.dot(xvx, xvy))
        py = np.multiply(vmat, y_xb)
        # add_py p_add_py
        add_py = np.multiply(gmat_eigenvals, py)
        xvy = np.dot(vx.T, add_py)
        y_xb = add_py - np.dot(xmat_trans, np.dot(xvx, xvy))
        p_add_py = np.multiply(vmat, y_xb)
        # res_py p_res_py
        res_py = py.copy()
        xvy = np.dot(vx.T, res_py)
        y_xb = res_py - np.dot(xmat_trans, np.dot(xvx, xvy))
        p_res_py = np.multiply(vmat, y_xb)
        # fd
        tr_vd = np.sum(np.multiply(vmat, gmat_eigenvals))
        xvdvx = np.dot(xmat_trans.T, vmat * gmat_eigenvals * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, add_py))
        fd_mat[0] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        tr_vd = np.sum(vmat)
        xvdvx = np.dot(xmat_trans.T, vmat * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, res_py))
        fd_mat[1] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        # AI
        ai_mat[0, 0] = np.sum(np.dot(add_py.T, p_add_py))
        ai_mat[0, 1] = ai_mat[1, 0] = np.sum(np.dot(add_py.T, p_res_py))
        ai_mat[1, 1] = np.sum(np.dot(res_py.T, p_res_py))
        ai_mat = 0.5 * ai_mat
        # EM
        em_mat[0, 0] = num_id / (2 * var[0] * var[0])
        em_mat[1, 1] = num_id / (2 * var[1] * var[1])
        for j in range(0, 51):
            gamma = j * 0.02
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_update = var + delta
            if min(var_update) > 0:
                logging.info('EM weight value: {}'.format(gamma))
                break
        # Convergence criteria
        logging.info('Updated variances: {}'.format(var_update))
        cc_val = np.sum(pow(delta, 2)) / np.sum(pow(var_update, 2))
        cc_val = np.sqrt(cc_val)
        var = var_update.copy()
        logging.info("CC: {}".format(cc_val))
        if cc_val < cc:
            break
    if cc_val < cc:
        convergence = True
    return var, convergence


def _weight_emai_eigen_snp(y_trans, xmat_trans, gmat_eigenvals, var, maxiter=20, cc=1.0e-8):
    fd_mat = np.zeros(2)
    ai_mat = np.zeros((2, 2))
    em_mat = np.zeros((2, 2))
    num_id = gmat_eigenvals.shape[0]
    var_update = var * 1000
    delta = var_update - var
    cc_val = 1000.0
    for i in range(maxiter):
        # logging.info("Iteration: {}".format(i+1))
        vmat = 1.0 / (gmat_eigenvals * var[0] + var[1])
        vx = np.multiply(vmat, xmat_trans)
        xvx = np.dot(xmat_trans.T, vx)
        xvx = linalg.inv(xvx)
        # py
        xvy = np.dot(vx.T, y_trans)
        y_xb = y_trans - np.dot(xmat_trans, np.dot(xvx, xvy))
        py = np.multiply(vmat, y_xb)
        # add_py p_add_py
        add_py = np.multiply(gmat_eigenvals, py)
        xvy = np.dot(vx.T, add_py)
        y_xb = add_py - np.dot(xmat_trans, np.dot(xvx, xvy))
        p_add_py = np.multiply(vmat, y_xb)
        # res_py p_res_py
        res_py = py.copy()
        xvy = np.dot(vx.T, res_py)
        y_xb = res_py - np.dot(xmat_trans, np.dot(xvx, xvy))
        p_res_py = np.multiply(vmat, y_xb)
        # fd
        tr_vd = np.sum(np.multiply(vmat, gmat_eigenvals))
        xvdvx = np.dot(xmat_trans.T, vmat * gmat_eigenvals * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, add_py))
        fd_mat[0] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        tr_vd = np.sum(vmat)
        xvdvx = np.dot(xmat_trans.T, vmat * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, res_py))
        fd_mat[1] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        # AI
        ai_mat[0, 0] = np.sum(np.dot(add_py.T, p_add_py))
        ai_mat[0, 1] = ai_mat[1, 0] = np.sum(np.dot(add_py.T, p_res_py))
        ai_mat[1, 1] = np.sum(np.dot(res_py.T, p_res_py))
        ai_mat = 0.5 * ai_mat
        # EM
        em_mat[0, 0] = num_id / (2 * var[0] * var[0])
        em_mat[1, 1] = num_id / (2 * var[1] * var[1])
        for j in range(0, 51):
            gamma = j * 0.02
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_update = var + delta
            if min(var_update) > 0:
                # logging.info('EM weight value: {}'.format(gamma))
                break
        # Convergence criteria
        # logging.info('Updated variances: {}'.format(var_update))
        cc_val = np.sum(pow(delta, 2)) / np.sum(pow(var_update, 2))
        cc_val = np.sqrt(cc_val)
        var = var_update.copy()
        # logging.info("CC: {}".format(cc_val))
        if cc_val < cc:
            break
    vmat = 1.0 / (gmat_eigenvals * var[0] + var[1])
    vx = np.multiply(vmat, xmat_trans)
    xvx = np.dot(xmat_trans.T, vx)
    xvx = linalg.inv(xvx)
    # py
    xvy = np.dot(vx.T, y_trans)
    b_hat = np.dot(xvx, xvy)
    snp_eff = b_hat[-1, -1]
    se = np.sqrt(xvx[-1, -1])
    t_val = snp_eff / se
    p_val = chi2.sf(t_val*t_val, 1)
    return cc_val, snp_eff, se, p_val



if __name__=='__main__':
    import logging
    import os
    from gmat.gmatrix import agmat
    logging.basicConfig(level=logging.INFO)
    pheno_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/pheno'
    bed_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/plink'
    # id-id-value form
    agmat_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/test'
    agmat2 = agmat(bed_file, out_file=agmat_file, inv=False, small_val=0.001, out_fmt='id_id_val')
    fixed(pheno_file, bed_file, agmat_file,
               out_file='/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/res.txt', npart=1)



import logging
import numpy as np
import pandas as pd
from scipy import linalg
from pysnptools.snpreader import Bed
from tqdm import tqdm
from gmat.process_plink.process_plink import impute_geno
from scipy.stats import chi2
from functools import reduce
from scipy.optimize import minimize


def fixed_min(pheno_file, bed_file, agmat_file, out_file='res.txt', npart=20):
    """
    run GWAS based on linear mixed model and put one snp in each analysis.
    :param pheno_file: phenotypic file. The fist column is individual id . The second column is always a column of 1s
    for population mean. The last column is phenotypic values. The other covariates can be added between columns for
    population mean and phenotypic values. Missing values in the phenotypic file can be
    expressed as 'NA', 'NaN', 'nan' or 'na'.
    :param bed_file: the prefix for the plink binary file.
    :param agmat_file: the prefix for the additive genomic relationship matrix file.
    :param out_file: the output file. Default is 'res.txt'.
    :param npart: divide the SNPs into n parts. Default is 20.
    :return:
    """
    logging.info("###Prepare the phenotypic values, covariates and genomic matrix")
    y, xmat, ag, id_keep = _pre_pheno(pheno_file, agmat_file)
    logging.info("###Estimate the variances in the null model.")
    init = np.array([np.var(y) / 2] * 2)
    gmat_eigenvals, gmat_eigenvecs = linalg.eigh(ag)
    gmat_eigenvals = np.array(gmat_eigenvals).reshape(-1, 1)
    y_trans = np.dot(gmat_eigenvecs.T, y)  # transform the phenotypic values
    xmat_trans = np.dot(gmat_eigenvecs.T, xmat)  # transform the fixed effects
    n = y_trans.shape[0]
    c = xmat_trans.shape[1]
    const = -(n - c) / 2 + (n - c) / 2 * np.log((n - c) / (2 * np.pi)) + \
            0.5 * np.linalg.slogdet(np.dot(xmat_trans.T, xmat_trans))[1]

    def opt_min(gamma):
        hinv = gmat_eigenvals*gamma[0] + 1
        logh = np.sum(np.log(hinv))
        hinv = 1/hinv
        hinv = hinv.reshape(1, -1)
        xmat_hinv = np.multiply(xmat_trans.T, hinv)
        xvx = np.dot(xmat_hinv, xmat_trans)
        logxvx = np.linalg.slogdet(xvx)[1]
        xvx = linalg.inv(xvx)
        pymat = np.multiply(hinv.T, y_trans) - np.dot(xmat_hinv.T, np.dot(xvx, np.dot(xmat_hinv, y_trans)))
        ypy = np.sum(np.dot(y_trans.T, pymat))
        logypy = np.log(ypy)
        lrt = const - (n-c)/2*logypy - 0.5*logh - 0.5*logxvx
        return -2*lrt

    gamma0 = init[0] / init[1]
    res = minimize(opt_min, np.array([gamma0]), method="L-BFGS-B", bounds=((0, None),),
                   options={'ftol': 1.0e-8, 'gtol': 1.0e-5, 'disp': True})
    logging.info("The information for the optimization: {}".format(res))
    convergence = res.success
    gamma = res.x[0]
    hinv = gmat_eigenvals * gamma + 1
    hinv = 1 / hinv
    hinv = hinv.reshape(1, -1)
    xmat_hinv = np.multiply(xmat_trans.T, hinv)
    xvx = np.dot(xmat_hinv, xmat_trans)
    xvx = linalg.inv(xvx)
    pymat = np.multiply(hinv.T, y_trans) - np.dot(xmat_hinv.T, np.dot(xvx, np.dot(xmat_hinv, y_trans)))
    ypy = np.sum(np.dot(y_trans.T, pymat))
    error_var = ypy/(n-c)
    add_var = gamma*error_var
    var = np.array([add_var, error_var])
    logging.info('Whether converge?: {}'.format(convergence))
    logging.info('The estimated variances: {}'.format(var))
    snp_on_disk = Bed(bed_file, count_A1=False)
    num_snp = snp_on_disk.sid_count
    fam_df = pd.read_csv(bed_file + '.fam', header=None, sep='\s+')
    id_in_fam = list(np.array(fam_df.iloc[:, 1], dtype=np.str))
    id_keep_ind = []
    for val in id_keep:
        id_keep_ind.append(id_in_fam.index(val))
    snp_index = []
    for i in range(npart):
        snp_index.append(int(num_snp / npart) * i)
    snp_index.append(num_snp)
    logging.info('###Start the association')
    with open(out_file, 'w') as fout:
        fout.write('\t'.join(['chro', 'snp_ID', 'pos', 'allele1', 'allele2', 'cc', 'eff', 'se', 'p_wald']))
        fout.write('\n')
    maxiter = 100
    if not convergence:
        maxiter = 0
    for i in range(npart):
        logging.info('The {}/{} part'.format(i+1, npart))
        snp_info = pd.read_csv(bed_file + '.bim', sep='\s+', header=None, skiprows=snp_index[i],
                               nrows=snp_index[i+1]-snp_index[i])
        res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
        res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
        snp_mat = snp_on_disk[id_keep_ind, snp_index[i]:snp_index[i + 1]].read().val
        if np.any(np.isnan(snp_mat)):
            snp_mat = impute_geno(snp_mat)
        cc_vec = []
        eff_vec = []
        se_vec = []
        p_vec = []
        for k in tqdm(range(snp_mat.shape[1])):
            snpk = np.dot(gmat_eigenvecs.T, snp_mat[:, k:(k+1)])
            xmat_trans_k = np.concatenate([xmat_trans, snpk], axis=1)
            cc_val, snp_eff, se, p_val = _opt_min_snp(y_trans, xmat_trans_k, gmat_eigenvals, var)
            cc_vec.append(cc_val)
            eff_vec.append(snp_eff)
            se_vec.append(se)
            p_vec.append(p_val)
        res_df.loc[:, 'cc_val'] = cc_vec
        res_df.loc[:, 'eff_val'] = eff_vec
        res_df.loc[:, 'se_val'] = se_vec
        res_df.loc[:, 'p_val'] = p_vec
        res_df.to_csv(out_file, sep='\t', index=False, header=False, mode='a')
    return 0


def _pre_pheno(pheno_file, agmat_file):
    id_in_agmat = []
    with open(agmat_file + '.id') as fin:
        for line in fin:
            arr = line.split()
            id_in_agmat.append(arr[0])
    y = []
    xmat = []
    id_keep = []  # keep the id both in the agmat file and pheno file, without missing phenotypes and covariates.
    with open(pheno_file) as fin:
        for line in fin:
            arr = line.split()
            if arr[0] in id_in_agmat:
                try:
                    vec = np.array(arr[1:], dtype=float)
                    y.append(vec[-1])
                    xmat.append(vec[:-1])
                    id_keep.append(arr[0])
                except Exception as e:
                    del e
    y = np.array(y).reshape(-1, 1)
    xmat = np.array(xmat)
    id_dct = dict(zip(id_keep, range(len(id_keep))))
    ag = np.zeros((len(id_keep), len(id_keep)))
    with open(agmat_file + '.agrm.id_fmt') as fin:
        for line in fin:
            arr = line.split()
            try:
                ag[id_dct[arr[0]], id_dct[arr[1]]] = ag[id_dct[arr[1]], id_dct[arr[0]]] = float(arr[2])
            except Exception as e:
                del e
    return y, xmat, ag, id_keep


def _opt_min_snp(y_trans, xmat_trans, gmat_eigenvals, var):
    n = y_trans.shape[0]
    c = xmat_trans.shape[1]
    const = -(n - c) / 2 + (n - c) / 2 * np.log((n - c) / (2 * np.pi)) + \
            0.5 * np.linalg.slogdet(np.dot(xmat_trans.T, xmat_trans))[1]

    def opt_min_snp(gamma):
        hinv = gmat_eigenvals * gamma[0] + 1
        logh = np.sum(np.log(hinv))
        hinv = 1 / hinv
        hinv = hinv.reshape(1, -1)
        xmat_hinv = np.multiply(xmat_trans.T, hinv)
        xvx = np.dot(xmat_hinv, xmat_trans)
        logxvx = np.linalg.slogdet(xvx)[1]
        xvx = linalg.inv(xvx)
        pymat = np.multiply(hinv.T, y_trans) - np.dot(xmat_hinv.T, np.dot(xvx, np.dot(xmat_hinv, y_trans)))
        ypy = np.sum(np.dot(y_trans.T, pymat))
        logypy = np.log(ypy)
        lrt = const - (n - c) / 2 * logypy - 0.5 * logh - 0.5 * logxvx
        return -2 * lrt

    gamma0 = var[0] / var[1]
    res = minimize(opt_min_snp, np.array([gamma0]), method="L-BFGS-B", bounds=((0, None),),
             options={'ftol': 1.0e-8, 'gtol': 1.0e-5, 'disp': False})
    gamma = res.x[0]
    hinv = gmat_eigenvals * gamma + 1
    hinv = 1 / hinv
    hinv = hinv.reshape(1, -1)
    xmat_hinv = np.multiply(xmat_trans.T, hinv)
    xvy = np.dot(xmat_hinv, y_trans)
    xvx = np.dot(xmat_hinv, xmat_trans)
    xvx = linalg.inv(xvx)
    pymat = np.multiply(hinv.T, y_trans) - np.dot(xmat_hinv.T, np.dot(xvx, np.dot(xmat_hinv, y_trans)))
    ypy = np.sum(np.dot(y_trans.T, pymat))
    error_var = ypy / (n - c)
    b_hat = np.dot(xvx, xvy)
    snp_eff = b_hat[-1, -1]
    se = np.sqrt(xvx[-1, -1]*error_var)
    t_val = snp_eff / se
    p_val = chi2.sf(t_val*t_val, 1)
    cc_val = 1000.0
    return cc_val, snp_eff, se, p_val


def _opt_gamma_snp(y_trans, xmat_trans, gmat_eigenvals, var):
    n = y_trans.shape[0]
    c = xmat_trans.shape[1]
    const = -(n - c) / 2 + (n - c) / 2 * np.log((n - c) / (2 * np.pi)) + \
            0.5 * np.linalg.slogdet(np.dot(xmat_trans.T, xmat_trans))[1]

    def opt_min_snp(gamma):
        hinv = gmat_eigenvals * gamma[0] + 1
        logh = np.sum(np.log(hinv))
        hinv = 1 / hinv
        hinv = hinv.reshape(1, -1)
        xmat_hinv = np.multiply(xmat_trans.T, hinv)
        xvx = np.dot(xmat_hinv, xmat_trans)
        logxvx = np.linalg.slogdet(xvx)[1]
        xvx = linalg.inv(xvx)
        pymat = np.multiply(hinv.T, y_trans) - np.dot(xmat_hinv.T, np.dot(xvx, np.dot(xmat_hinv, y_trans)))
        ypy = np.sum(np.dot(y_trans.T, pymat))
        logypy = np.log(ypy)
        lrt = const - (n - c) / 2 * logypy - 0.5 * logh - 0.5 * logxvx
        return -2 * lrt

    gamma0 = var[0] / var[1]
    res = minimize(opt_min_snp, np.array([gamma0]), method="L-BFGS-B", bounds=((0, None),),
             options={'ftol': 1.0e-8, 'gtol': 1.0e-5, 'disp': False})
    gamma = res.x[0]
    hinv = gmat_eigenvals * gamma + 1
    hinv = 1 / hinv
    hinv = hinv.reshape(1, -1)
    xmat_hinv = np.multiply(xmat_trans.T, hinv)
    xvy = np.dot(xmat_hinv, y_trans)
    xvx = np.dot(xmat_hinv, xmat_trans)
    xvx = linalg.inv(xvx)
    pymat = np.multiply(hinv.T, y_trans) - np.dot(xmat_hinv.T, np.dot(xvx, np.dot(xmat_hinv, y_trans)))
    ypy = np.sum(np.dot(y_trans.T, pymat))
    error_var = ypy / (n - c)
    b_hat = np.dot(xvx, xvy)
    snp_eff = b_hat[-1, -1]
    se = np.sqrt(xvx[-1, -1]*error_var)
    t_val = snp_eff / se
    p_val = chi2.sf(t_val*t_val, 1)
    cc_val = 1000.0
    return cc_val, snp_eff, se, p_val



if __name__=='__main__':
    import logging
    import os
    from gmat.gmatrix import agmat
    logging.basicConfig(level=logging.INFO)
    pheno_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/pheno'
    bed_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/plink'
    # id-id-value form
    agmat_file = '/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/test'
    agmat2 = agmat(bed_file, out_file=agmat_file, inv=False, small_val=0.001, out_fmt='id_id_val')
    fixed(pheno_file, bed_file, agmat_file,
               out_file='/nfs1/ningc/WORK/Acode/GMAT/PKG/examples/data/mouse/res.txt', npart=1)



import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, hstack, block_diag
from scipy.stats import chi2
import datetime
import gc
import logging
from tqdm import tqdm
from patsy import dmatrix
from functools import reduce
import sympy


from gmat.process_plink.process_plink import impute_geno
from pysnptools.snpreader import Bed
from ._pre_data import _pre_data
from ._common import _design_matrix, _leg_mt, _full_col_rank


def unbalance_longwas_fixed(data_file, id, tpoint, trait, bed_file, agmat_file, var_com, condition_snp=None, snp_lst=None,
            tfix=None, fix=None, forder=3, aorder=3, porder=3, na_method='omit',
                             prefix_outfile='unbalance_longwas_fixed'):
    """
    the longitudinal GWAS for the unbalanced data treating the SNP as the time varied fixed effect.
    :param data_file: the data file. The first row is the variate names whose first initial position is alphabetical.
    For the class variates, the first letter must be capital; for the covariates (continuous variates), the first letter
    must be lowercase.
    :param id: A class variate name which indicates the individual id column in the data file.
    :param tpoint: A covariate names which indicates the time point column in the data file.
    :param trait: A variate name which indicates the analyzed trait column in the data file.
    :param bed_file: the prefix for the plink binary file.
    :param agmat_file: the file for genomic relationship matrix. This file can be produced by
    gmat.gmatrix.agmat function using agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')
    :param var_com: the estimated variance parameters by the gmat.longwas.unbalance.unbalance_varcom function.
    :param condition_snp: SNP ID or a list of SNP ID for conditional snps added to the time fixed effect.
    :param snp_lst: the snp list to test. Default is None.
    :param tfix: A class variate name for the time varied fixed effect. Default value is None. Only one time varied
    fixed effect can be included in the current version.
    :param fix: Expression for the time independent fixed effect. Default value is None. An example:
    fix = "Sex + age + Season".
    :param forder: the order of Legendre polynomials for the time varied fixed effect. The default value is 3.
    :param aorder: the order of Legendre polynomials for the additive genetic effect. The default value is 3.
    :param porder: the order of Legendre polynomials for the permanent environment effect. The default value is 3.
    :param na_method: The method to deal with missing values. The default value is 'omit'. 'omit' method will delete the
    row with missing values. 'include' method will fill the missing values with the adjacent values.
    :param prefix_outfile: the prefix for the output file. Default is 'unbalance_longwas_fixed'.
    :return: A pandas data frame for the test result.
    """
    logging.info('################################')
    logging.info('###Prepare the related matrix###')
    logging.info('################################')
    if var_com.shape[0] != aorder*(aorder+1)/2 + aorder + 1 + porder*(porder+1)/2 + porder + 1 + 1:
        logging.error('ERROR: Variances do not match the data, please check')
        sys.exit()
    data_df, ag, id_code, id_order = _pre_data(data_file, id, tpoint, trait, agmat_file, na_method=na_method)
    logging.info('***Build the design matrix for fixed effect***')
    logging.info('Time dependent fixed effect: {}'.format(tfix))
    tmin = np.min(data_df[tpoint])
    tmax = np.max(data_df[tpoint])
    xmat_t = _leg_mt(data_df[tpoint], tmax, tmin, forder)
    xmat_t = np.concatenate(xmat_t, axis=1)
    if tfix is not None:
        mat = []
        dmat_t = _design_matrix(tfix, data_df)
        for i in range(dmat_t.shape[1]):
            mat.append(xmat_t * dmat_t[:, i:(i + 1)])
        xmat_t = np.concatenate(mat, axis=1)
        del mat, dmat_t
        gc.collect()
    logging.info('Time independent fixed effect: {}'.format(fix))
    xmat = xmat_t
    if fix is not None:
        xmat_nt = _design_matrix(fix, data_df)
        xmat = np.concatenate([xmat_t, xmat_nt[:, 1:]], axis=1)
    if np.linalg.matrix_rank(xmat) < xmat.shape[1]:
        xmat = _full_col_rank(xmat)
    logging.info("Conditional SNP: {}".format(condition_snp))
    snp_on_disk = Bed(bed_file, count_A1=False)
    fam_df = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id_geno = list(np.array(fam_df.iloc[:, 1], dtype=str))
    id_order_index = []
    for i in id_order:
        id_order_index.append(id_geno.index(i))
    leg_lst = []  # legendre polynomials for time dependent fixed SNP effects, save for each individuals
    for val in id_order:
        leg_val = _leg_mt(data_df[data_df[id] == val][tpoint], tmax, tmin, forder)
        leg_val = np.concatenate(leg_val, axis=1)
        leg_lst.append(leg_val)
    if condition_snp is not None:
        if not isinstance(condition_snp, list):
            condition_snp = [condition_snp]
        condition_snp_index = snp_on_disk.sid_to_index(condition_snp)
        condition_snp_val = snp_on_disk[id_order_index, condition_snp_index].read().val
        if np.any(np.isnan(condition_snp_val)):
            logging.info('Missing genotypes of conditional SNP are imputed with random genotypes.')
            condition_snp_val = impute_geno(condition_snp_val)
        for i in range(condition_snp_val.shape[1]):
            snp_condition = list(map(lambda x, y: x * y, leg_lst, list(condition_snp_val[:, i])))
            snp_condition = np.concatenate(snp_condition, axis=0)
            xmat = np.concatenate((xmat, snp_condition), axis=1)
    logging.info('***Build the dedign matrix for random effect***')
    logging.info('Legendre order for additive effects: {}'.format(aorder))
    leg_add = _leg_mt(data_df[tpoint], tmax, tmin, aorder)
    row = np.array(range(data_df.shape[0]))
    col = np.array(id_code)
    val = np.array([1.0] * data_df.shape[0])
    add_mat = csr_matrix((val, (row, col)), shape=(data_df.shape[0], ag.shape[0]))
    zmat_add = []
    for i in range(len(leg_add)):
        zmat_add.append(add_mat.multiply(leg_add[i]))
    logging.info('Legendre order for permanent environmental effect: {}'.format(porder))
    leg_per = _leg_mt(data_df[tpoint], tmax, tmin, porder)
    per_mat = csr_matrix((val, (row, col)))
    zmat_per = []
    for i in range(len(leg_per)):
        zmat_per.append((per_mat.multiply(leg_per[i])))
    del row, col, val, leg_add, add_mat, leg_per, per_mat
    gc.collect()
    zmat = [zmat_add, zmat_per]
    y = np.array(data_df[trait]).reshape(-1, 1)

    logging.info('***Calculate the phenotypic (co)variance and P matrix***')
    eff_ind = [[0, xmat.shape[1]]]
    for i in range(len(zmat)):
        temp = [eff_ind[i][-1]]
        for j in range(len(zmat[i])):
            temp.append(temp[-1] + zmat[i][j].shape[1])
        eff_ind.append(temp)
    zmat_con_lst = [hstack(zmat[0]), hstack(zmat[1])]  # combined random matrix

    def cal_cov(var_ind, var_com):
        """
        return the matrix format of covariance.
        """
        cov_df = var_com.loc[var_com.loc[:, 'var']==var_ind, :]
        row = np.array(cov_df['vari']) - 1
        col = np.array(cov_df['varj']) - 1
        val = cov_df['var_com']
        cov_mat = csr_matrix((val, (row, col))).toarray()
        cov_mat = cov_mat + np.tril(cov_mat, k=-1).T
        return cov_mat
    add_cov = cal_cov(1, var_com)
    per_cov = cal_cov(2, var_com)
    res_var = np.array(var_com['var_com'])[-1]
    vmat = zmat_con_lst[0].dot((zmat_con_lst[0].dot(np.kron(add_cov, ag))).T)
    one_id = sparse.eye(zmat_con_lst[1].shape[1]/per_cov.shape[0])
    vmat = vmat + zmat_con_lst[1].dot((zmat_con_lst[1].dot(sparse.kron(per_cov, one_id))).T)
    vmat_diag = np.diag(vmat) + res_var
    np.fill_diagonal(vmat, vmat_diag)
    vmat = linalg.inv(vmat)
    xv = np.dot(xmat.T, vmat)
    xvx = np.dot(xv, xmat)
    xvx = linalg.inv(xvx)
    pmat = vmat - reduce(np.dot, [xv.T, xvx, xv])
    logging.info('***Read the snp data***')
    num_id = snp_on_disk.iid_count
    num_snp = snp_on_disk.sid_count
    logging.info("There are {:d} individuals and {:d} SNPs.".format(num_id, num_snp))
    if snp_lst is None:
        snp_lst = range(num_snp)
    snp_lst = list(snp_lst)
    if min(snp_lst) < 0 or max(snp_lst) >= num_snp:
        logging.info('The value in the snp list should be >= {} and < {}', 0, num_snp)
        exit()
    snp_mat = snp_on_disk[id_order_index, snp_lst].read().val
    if np.any(np.isnan(snp_mat)):
        logging.info('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    logging.info('#####################################################################')
    logging.info('###Start the fixed regression longitudinal GWAS for unbalance data###')
    logging.info('#####################################################################')
    tpoint_vec = sorted(set(data_df[tpoint]))
    leg_tpoint_mat = _leg_mt(np.array(tpoint_vec), tmax, tmin, forder)
    leg_tpoint_mat = np.concatenate(leg_tpoint_mat, axis=1)
    leg_tpoint_accum = np.sum(leg_tpoint_mat, axis=0)
    chi_df = leg_lst[0].shape[1]
    eff_main_vec = []
    se_main_vec = []
    p_main_vec = []
    eff_vec = []
    chi_vec = []
    p_gbye_vec = []
    p_vec = []
    p_min_vec = []
    eff_accum_vec = []
    p_accum_vec = []
    eff_tpoint_mat = []
    p_tpoint_mat = []
    # eye_forder = sparse.eye(forder + 1, format="csr")
    for i in tqdm(range(snp_mat.shape[1])):
        snp_fix = list(map(lambda x, y: x*y, leg_lst, list(snp_mat[:, i])))
        snp_fix = np.concatenate(snp_fix, axis=0)
        # snp_fix = leg_mat.dot((sparse.kron(snp_mat[:, i:(i+1)], eye_forder)))
        # snp_fix = snp_fix.toarray()
        p_xsnp = np.dot(pmat, snp_fix[:, 0:1])
        xpx = np.dot(snp_fix[:, 0:1].T, p_xsnp)
        se_main = np.sqrt(xpx[0, 0])
        xpy = np.dot(p_xsnp.T, y)
        eff_main = xpy[-1, -1]
        chi_val = eff_main*eff_main/(se_main*se_main)
        p_main = chi2.sf(chi_val, 1)
        eff_main_vec.append(eff_main)
        se_main_vec.append(se_main)
        p_main_vec.append(p_main)
        p_xsnp = np.dot(pmat, snp_fix)
        xpx = np.dot(snp_fix.T, p_xsnp)
        xpx = linalg.inv(xpx)
        xpy = np.dot(p_xsnp.T, y)
        b = np.dot(xpx, xpy)
        eff = b[-(chi_df-1):, -1]
        eff_var = xpx[-(chi_df-1):, -(chi_df-1):]
        chi_val = np.sum(np.dot(np.dot(eff.T, linalg.inv(eff_var)), eff))
        p_gbye = chi2.sf(chi_val, chi_df-1)
        p_gbye_vec.append(p_gbye)
        eff = b[-chi_df:, -1]
        eff_var = xpx[-chi_df:, -chi_df:]
        chi_val = np.sum(np.dot(np.dot(eff.T, linalg.inv(eff_var)), eff))
        p_val = chi2.sf(chi_val, chi_df)
        eff_vec.append(eff)
        chi_vec.append(chi_val)
        p_vec.append(p_val)
        eff_tpoint_vec = []
        p_tpoint_vec = []
        for k in range(leg_tpoint_mat.shape[0]):
            eff_tpoint = np.sum(np.dot(leg_tpoint_mat[k, :], eff))
            eff_tpoint_vec.append(eff_tpoint)
            eff_var_tpoint = np.sum(np.dot(leg_tpoint_mat[k, :], np.dot(eff_var, leg_tpoint_mat[k, :])))
            chi_tpoint = eff_tpoint*eff_tpoint/eff_var_tpoint
            p_tpoint = chi2.sf(chi_tpoint, 1)
            p_tpoint_vec.append(p_tpoint)
        eff_tpoint_mat.append(np.array(eff_tpoint_vec).reshape(1, -1))
        p_tpoint_mat.append(np.array(p_tpoint_vec).reshape(1, -1))
        p_min_vec.append(min(p_tpoint_vec))
        eff_accum = np.sum(np.dot(leg_tpoint_accum, eff))
        eff_accum_vec.append(eff_accum)
        eff_var_accum = np.sum(np.dot(leg_tpoint_accum, np.dot(eff_var, leg_tpoint_accum)))
        chi_accum = eff_accum*eff_accum/eff_var_accum
        p_accum = chi2.sf(chi_accum, 1)
        p_accum_vec.append(p_accum)
    logging.info('Finish association analysis')
    logging.info('***Output***')
    head_tpoint = []
    for val in tpoint_vec:
        head_tpoint.append('eff_tpoint_' + str(val))
    for val in tpoint_vec:
        head_tpoint.append('p_tpoint_' + str(val))
    eff_tpoint_mat = np.concatenate(eff_tpoint_mat, axis=0)
    p_tpoint_mat = np.concatenate(p_tpoint_mat, axis=0)
    res_tpoint = np.concatenate([eff_tpoint_mat, p_tpoint_mat], axis=1)
    np.savetxt(prefix_outfile + '.tpoint', res_tpoint, header='\t'.join(head_tpoint), delimiter="\t", comments='')
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[snp_lst, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'order'] = snp_lst
    res_df = res_df.iloc[:, [5, 0, 1, 2, 3, 4]]
    res_df.loc[:, 'eff_main'] = eff_main_vec
    res_df.loc[:, 'se_main'] = se_main_vec
    res_df.loc[:, 'p_main'] = p_main_vec
    eff_vec = np.array(eff_vec)
    for i in range(eff_vec.shape[1]):
        col_ind = 'eff' + str(i)
        res_df.loc[:, col_ind] = eff_vec[:, i]
    res_df.loc[:, 'eff_accum'] = eff_accum_vec
    res_df.loc[:, 'chi_val'] = chi_vec
    res_df.loc[:, 'p_gbye'] = p_gbye_vec
    res_df.loc[:, 'p_val'] = p_vec
    res_df.loc[:, 'p_min'] = p_min_vec
    res_df.loc[:, 'p_accum'] = p_accum_vec
    out_file = prefix_outfile + '.res'
    res_df.to_csv(out_file, sep='\t', index=False)
    return res_df



import logging
from scipy import sparse
from scipy.sparse import csr_matrix, hstack
import numpy as np
from scipy import linalg
import pandas as pd
from patsy import dmatrix
import sympy
import gc
import os

from ._pre_data import _pre_data
from ._common import _design_matrix, _leg_mt, _full_col_rank
from ._emai import _emai


def unbalance_varcom(data_file, id, tpoint, trait, agmat_file, tfix=None, fix=None, forder=3, aorder=3, porder=3,
             na_method='omit', init=None, maxiter=100, cc_par=1.0e-8, cc_gra=1.0e-3,
                     em_weight_step=0.01, out_file='unbalance_varcom'):
    """
    Estimate the variance parameters for the unbalance longitudinal data using the random regression model
    :param data_file: the data file. The first row is the variate names whose first initial position is alphabetical.
    For the class variates, the first letter must be capital; for the covariates (continuous variates), the first letter
    must be lowercase.
    :param id: A class variate name which indicates the individual id column in the data file.
    :param tpoint: A covariate names which indicates the time point column in the data file.
    :param trait: A variate name which indicates the analyzed trait column in the data file.
    :param agmat_file: the prefix for genomic relationship matrix file. This file can be produced by
    gmat.gmatrix.agmat function using agmat(bed_file, out_fmt='id_id_val')
    :param tfix: variate names for the time varied fixed effect. Default value is None. An example:
    tfix = "Sex + age + Season".
    :param fix: Expression for the time independent fixed effect. Default value is None. An example:
    fix = "Sex + age + Season".
    :param forder: the order of Legendre polynomials for the time varied fixed effect. The default value is 3.
    :param aorder: the order of Legendre polynomials for the additive genetic effect. The default value is 3.
    :param porder: the order of Legendre polynomials for the permanent environment effect. The default value is 3.
    :param na_method: The method to deal with missing values. The default value is 'omit'. 'omit' method will delete the
    row with missing values. 'include' method will fill the missing values with the adjacent values.
    :param init: the initial values for the variance parameters. The default value is None.
    :param maxiter: the maximum number of iteration. Default is 100.
    :param cc_par: Convergence criteria for the changed variance parameters. Default is 1.0e-8.
    :param cc_gra: Convergence criteria for the norm of gradient vector. Default is 1.0e-3.
    :param em_weight_step: the step of the em weight. Default is 0.001.
    :param out_file: the prefix for the output file. Default is 'unbalance_varcom'.
    :return: the estimated variance parameters.
    """
    logging.info('########################################################################')
    logging.info('###Prepare the data for unbalanced longitudinal variances estimation.###')
    logging.info('########################################################################')
    logging.info('***Read the data file and genomic relationship matrix***')
    data_df, ag, id_code, _ = _pre_data(data_file, id, tpoint, trait, agmat_file, na_method=na_method)
    ag_inv = linalg.inv(ag)
    del ag
    gc.collect()
    logging.info('***Build the design matrix for fixed effect***')
    logging.info('Time dependent fixed effect: {}'.format(tfix))
    tmin = np.min(data_df[tpoint])
    tmax = np.max(data_df[tpoint])
    xmat_t = _leg_mt(data_df[tpoint], tmax, tmin, forder)
    xmat_t = np.concatenate(xmat_t, axis=1)
    if tfix is not None:
        mat = []
        dmat_t = _design_matrix(tfix, data_df)
        for i in range(dmat_t.shape[1]):
            mat.append(xmat_t * dmat_t[:, i:(i+1)])
        xmat_t = np.concatenate(mat, axis=1)
        del mat, dmat_t
        gc.collect()
    logging.info('Time independent fixed effect: {}'.format(fix))
    xmat = xmat_t
    if fix is not None:
        xmat_nt = _design_matrix(fix, data_df)
        xmat = np.concatenate([xmat_t, xmat_nt[:, 1:]], axis=1)
    if np.linalg.matrix_rank(xmat) < xmat.shape[1]:
        xmat = _full_col_rank(xmat)
    logging.info('***Build the dedign matrix for random effect***')
    logging.info('Legendre order for additive effects: {}'.format(aorder))
    leg_add = _leg_mt(data_df[tpoint], tmax, tmin, aorder)
    row = np.array(range(data_df.shape[0]))
    col = np.array(id_code)
    val = np.array([1.0] * data_df.shape[0])
    add_mat = csr_matrix((val, (row, col)), shape=(data_df.shape[0], ag_inv.shape[0]))
    zmat_add = []
    for i in range(len(leg_add)):
        zmat_add.append(add_mat.multiply(leg_add[i]))
    logging.info('Legendre order for permanent environmental effect: {}'.format(porder))
    leg_per = _leg_mt(data_df[tpoint], tmax, tmin, porder)
    per_mat = csr_matrix((val, (row, col)))
    zmat_per = []
    for i in range(len(leg_per)):
        zmat_per.append((per_mat.multiply(leg_per[i])))
    del row, col, val, leg_add, add_mat, leg_per, per_mat
    gc.collect()
    zmat = [zmat_add, zmat_per]
    y = np.array(data_df[trait]).reshape(-1, 1)
    logging.info('###########################################################')
    logging.info('###Start the unbalanced longitudinal variances estimation###')
    logging.info('###########################################################')
    res = _emai(y, xmat, zmat, ag_inv, init=init, maxiter=maxiter, cc_par=cc_par, cc_gra=cc_gra, em_weight_step=em_weight_step,
          out_file=out_file)
    var_file = out_file + '.var'
    res.to_csv(var_file, sep='\t', index=False)
    return res



import numpy as np
import pandas as pd
import logging
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy.stats import beta
from tqdm import tqdm
import logging


def manplot(gwas_res_file, chro, pos, p, color_vec=("gray", "dimgray"), sig_color='red',
            threshold=1.0e-7, logp=True, out_file='man.png'):
    """
    Draw the Manhattan plot from GWAS result file.
    :param gwas_res_file: the GWAS result file with header
    :param chro: the column name for chromosome
    :param pos: the column name for position
    :param p: the column name for P values
    :param color_vec: color for different chromosomes.
    :param sig_color: the color for point pass the threshold. Default is 'red'.
    :param threshold: the threshold value. Default is 1.0e-7.
    :param logp: Whether to log transform the p values. Default is True.
    :param out_file: the output file. Default is "man.png".
    :return:
    """
    p_dct = {}
    pos_dct = {}
    chro_vec = []
    with open(gwas_res_file) as fin:
        head_line = fin.readline()
        arr = head_line.split()
        chro_ind = arr.index(chro)
        pos_ind = arr.index(pos)
        p_ind = arr.index(p)
        line_num = 0
        for line in fin:
            line_num += 1
            arr = line.split()
            try:
                pi = float(arr[p_ind])
            except Exception as e:
                del e
                logging.info('The p value in Line {} is not numerical.'.format(line_num))
                continue
            try:
                posi = float(arr[pos_ind])
            except Exception as e:
                del e
                logging.info('The position value in Line {} is not numerical.'.format(line_num))
                continue
            chroi = arr[chro_ind]
            if chroi not in chro_vec:
                chro_vec.append(chroi)
            try:
                p_dct[chroi].append(pi)
            except Exception as e:
                del e
                p_dct[chroi] = [pi]
            try:
                pos_dct[chroi].append(posi)
            except Exception as e:
                del e
                pos_dct[chroi] = [posi]
    plt.figure(figsize=(10, 3.5), dpi=600)
    plt.rc('font', family='Times New Roman', size=10)
    pos_accum = [0.0]
    xtick_vec = []
    for i in range(len(chro_vec)):
        posx = np.array(pos_dct[chro_vec[i]]) + pos_accum[-1]
        if logp:
            pvaly = -np.log10(p_dct[chro_vec[i]])
            threshold_line = -np.log10(threshold)
        else:
            pvaly = np.array(p_dct[chro_vec[i]])
            threshold_line = threshold
        plt.scatter(posx, pvaly, s=1.5, c=color_vec[i%len(color_vec)])
        p_line = np.array([threshold_line] * len(posx))
        plt.plot(posx, p_line, 'b')
        posx_sig = posx[pvaly > threshold_line]
        pvaly_sig = pvaly[pvaly > threshold_line]
        plt.scatter(posx_sig, pvaly_sig, s=1.5, c=sig_color)
        pos_accum.append(pos_accum[-1] + np.max(pos_dct[chro_vec[i]]))
        xtick_vec.append((pos_accum[-1] + pos_accum[-2])/2)
    plt.xticks(xtick_vec, chro_vec, fontsize=7)
    plt.savefig(out_file)
    return 0



import numpy as np
import pandas as pd


def _emai(y_lst, xmat_lst, zmat_lst, ag, error_lst, init=init, maxiter=maxiter, cc_par=cc_par,
                      cc_gra=cc_gra, em_weight_step=em_weight_step):
    def init_var(y_lst):
        """
        initialize the variances
        """
        y_var_lst = np.array(list(map(np.var, y_lst)))/2
        add_cov = np.diag(y_var_lst)
        add_ind = np.tril_indices_from(add_cov)
        error_cov = np.diag(y_var_lst)
        error_ind = np.tril_indices_from(error_cov)
        cov_ind = [1] * len(add_ind[0]) + [2] * len(error_ind[0])
        cov_ind_i = list(add_ind[0] + 1) + list(error_ind[0] + 1)
        cov_ind_j = list(add_ind[1] + 1) + list(error_ind[1] + 1)
        var_com = list(add_cov[add_ind]) + list(error_cov[error_ind])
        var_df = {'var': cov_ind,
                  "vari": cov_ind_i,
                  "varj": cov_ind_j,
                  "var_com": var_com}
        var_df = pd.DataFrame(var_df, columns=['var', "vari", "varj", "var_com"])
        return np.array(var_com), var_df
    var_com, var_df = init_var(y_lst)
    if init is not None:
        if len(var_com) != len(init):
            logging.error('The length of initial variances should be {}'.format(len(var_com)))
            sys.exit()
        else:
            var_com = np.array(init)
            var_df['var_com'] = var_com

   def pre_cov(y_lst, var_df):
        """
        return the covariance matrix. If one covariance matrix is not positive, return None.
        """
        add_cov = np.zeros((len(y_lst), len(y_lst)))
        add_cov[np.tril_indices(len(y_lst))] = np.array(var_df[var_df['var']==1].iloc[:, -1])
        add_cov = np.add(add_cov, np.tril(add_cov, -1).T)
        error_cov = np.zeros((len(y_lst), len(y_lst)))
        error_cov[np.tril_indices(len(y_lst))] = np.array(var_df[var_df['var'] == 2].iloc[:, -1])
        error_cov = np.add(error_cov, np.tril(error_cov, -1).T)
        try:
            linalg.cholesky(add_cov)
            linalg.cholesky(error_cov)
        except Exception as _:
                return None
        return add_cov, error_cov

    cov_lst = pre_cov(y_lst, var_df)
    if cov_lst is None:
        logging.error("ERROR: Initial variances is not positive define, please check!")
        sys.exit()
    iter_count = 0
    cc_par_val = 1000.0
    cc_gra_val = 1000.0
    delta = 1000.0
    var_com_update = var_com * 1000
    logging.info("initial variances: {}".format(var_df))
    while iter_count < maxiter:



import logging

from ._pre_data import _pre_data


def varcom(data_file, id, trait, agmat_file, fix=None, init=None, maxiter=100, cc_par=1.0e-8, cc_gra=1.0e-3,
                     em_weight_step=0.01, prefix_outfile='mvlmm_varcom'):
    """
    Estimate the variances for the multivariate linear mixed model.
    :param data_file:
    :param trait:
    :param agmat_file:
    :param fix:
    :param init:
    :param maxiter:
    :param cc_par:
    :param cc_gra:
    :param em_weight_step:
    :param prefix_outfile:
    :return:
    """
    logging.info("######################")
    logging.info("###Prepare the data###")
    logging.info("######################")
    y_lst, xmat_lst, zmat_lst, ag, error_lst = _pre_data(data_file, id, trait, agmat_file, fix=fix)
    varcom_df = _emai(y_lst, xmat_lst, zmat_lst, ag, error_lst, init=init, maxiter=maxiter, cc_par=cc_par,
                      cc_gra=cc_gra, em_weight_step=em_weight_step)



import sys
import logging
import numpy as np
from scipy import linalg
import pandas as pd
import time
from gmat.process_plink.process_plink import impute_geno
from pysnptools.snpreader import Bed
from tqdm import tqdm


def output_mat(mat, id_df, out_file, out_fmt):
    if out_fmt == 'mat':
        np.savetxt(out_file + '.mat_fmt', mat)
    elif out_fmt == 'row_col_val':
        ind = np.tril_indices_from(mat)
        df = pd.DataFrame({
            "row": ind[0]+1,
            "col": ind[1]+1,
            "val": mat[ind]
        })
        df.to_csv(out_file + '.ind_fmt', sep=' ', index=False, header=False, columns=["row", "col", "val"])
    elif out_fmt == 'id_id_val':
        id = np.array(id_df)
        ind = np.tril_indices_from(mat)
        df = pd.DataFrame({
            "id0": id[ind[0]],
            "id1": id[ind[1]],
            "val": mat[ind]
        })
        df.to_csv(out_file + '.id_fmt', sep=' ', index=False, header=False, columns=["id0", "id1", "val"])
    else:
        logging.error('The output format {} is not recognized, please check!'.format(out_fmt))
        sys.exit()
    return 1


def agmat(bed_file, out_file, inv=False, out_fmt='mat', npart=10, small_val=0.001):
    """
    additive genomic relationship matrix and its inversion
    :param bed_file: The prefix for plink binary file
    :param out_file: the prefix for the output file
    :param inv: Whether to calculate the inversion. Default value is True
    :param out_fmt: the output format. mat: matrix format (default); row_col_val: row-column-value format;
    id_id_val: id-id-value format.
    :param npart: the number of part. If memory is not enough, please increase the value.
    :param small_val: A small vale added to the diagonal to grant the positive definite. Default value is 0.001.
    :return: return numpy array for genomic relationship matrix and its inversion. Output the matrixes into the file
    with prefix of out_file.
    """
    logging.info("{:#^80}".format("Read the SNP and calculate"))
    snp_on_disk = Bed(bed_file, count_A1=False)
    num_snp = snp_on_disk.sid_count
    num_id = snp_on_disk.iid_count
    logging.info("There are {} individuals, {} SNPs.".format(num_id, num_snp))
    snp_index = []
    for i in range(npart):
        snp_index.append(int(num_snp/npart)*i)
    snp_index.append(num_snp)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    scale = 0.0
    kin = np.zeros((num_id, num_id))
    for i in tqdm(range(npart)):
        snp_mat = snp_on_disk[:, snp_index[i]:snp_index[i+1]].read().val
        if np.any(np.isnan(snp_mat)):
            snp_mat = impute_geno(snp_mat)
        freq = np.sum(snp_mat, axis=0) / (2 * num_id)
        freq.shape = (1, -1)
        scale += np.sum(2 * freq * (1 - freq))  # scale factor
        snp_mat = snp_mat - 2 * freq
        kin += np.dot(snp_mat, snp_mat.T)
    kin = kin / scale
    kin_diag = np.diag(kin)
    kin_diag = kin_diag + kin_diag * small_val
    np.fill_diagonal(kin, kin_diag)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    logging.info("{:#^80}".format("Output"))
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    logging.info("The prefix of output file is " + out_file)
    fam_info.iloc[:, 1].to_csv(out_file + '.id', index=False, header=False, sep=' ')
    output_mat(kin, fam_info.iloc[:, 1], out_file + '.agrm', out_fmt)
    kin_inv = None
    if inv:
        logging.info("{:#^80}".format("Calculate the inversion of kinship"))
        clock_t0 = time.perf_counter()
        cpu_t0 = time.process_time()
        kin_inv = linalg.inv(kin)
        clock_t1 = time.perf_counter()
        cpu_t1 = time.process_time()
        logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0,
                                                                                   cpu_t1 - cpu_t0))
        logging.info("{:#^80}".format("Output the inversion"))
        logging.info("The prefix of output file is: " + out_file)
        output_mat(kin_inv, fam_info.iloc[:, 1], out_file + '.agiv', out_fmt)
    return kin, kin_inv


def agmat_dosage(geno_file, out_file, inv=False, out_fmt='mat', npart=10, small_val=0.001):
    """
    additive genomic relationship matrix and its inversion using dosage genotype.
    :param geno_file: dosage genotype file. The first row is the individual id. The dosage genotype of each individual
    is saved from the second row.
    :param out_file: the prefix for the output file
    :param inv: Whether to calculate the inversion. Default value is True
    :param out_fmt: the output format. mat: matrix format (default); row_col_val: row-column-value format;
    id_id_val: id-id-value format.
    :param npart: the number of part. If memory is not enough, please increase the value.
    :param small_val: A small vale added to the diagonal to grant the positive definite. Default value is 0.001.
    :return: return numpy array for genomic relationship matrix and its inversion. Output the matrixes into the file
    with prefix of out_file.
    """
    logging.info("{:#^80}".format("Read the SNP and calculate"))
    num_snp = 0
    with open(geno_file, 'r') as fin:
        line = fin.readline()
        id = line.split()
        for _ in fin:
            num_snp += 1
    num_id = len(id)
    np.savetxt(out_file + '.id', np.array(id).reshape(-1, 1), fmt="%s")
    logging.info("There are {} individuals, {} SNPs.".format(num_id, num_snp))
    snp_index = []
    for i in range(npart):
        snp_index.append(int(num_snp/npart)*i)
    snp_index.append(num_snp)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    scale = 0.0
    kin = np.zeros((num_id, num_id))
    for i in tqdm(range(npart)):
        snp_mat = np.loadtxt(geno_file, skiprows=snp_index[i]+1, max_rows=snp_index[i+1]-snp_index[i])
        snp_mean = np.mean(snp_mat, axis=1).reshape(-1, 1)
        snp_var = np.var(snp_mat, axis=1)
        scale += np.sum(snp_var)  # scale factor
        snp_mat = snp_mat - snp_mean
        kin += np.dot(snp_mat.T, snp_mat)
    kin = kin / scale
    kin_diag = np.diag(kin)
    kin_diag = kin_diag + kin_diag * small_val
    np.fill_diagonal(kin, kin_diag)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    logging.info("{:#^80}".format("Output"))
    output_mat(kin, id, out_file + '.agrm', out_fmt)
    kin_inv = None
    if inv:
        logging.info("{:#^80}".format("Calculate the inversion of kinship"))
        clock_t0 = time.perf_counter()
        cpu_t0 = time.process_time()
        kin_inv = linalg.inv(kin)
        clock_t1 = time.perf_counter()
        cpu_t1 = time.process_time()
        logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0,
                                                                                   cpu_t1 - cpu_t0))
        logging.info("{:#^80}".format("Output the inversion"))
        logging.info("The prefix of output file is: " + out_file)
        output_mat(kin_inv, id, out_file + '.agiv', out_fmt)
    return kin, kin_inv


def dgmat_as(bed_file, out_file, inv=False, out_fmt='mat', npart=10, small_val=0.001):
    """
    dominance genomic relationship matrix and its inversion
    :param bed_file: The prefix for plink binary file
    :param out_file: the prefix for the output file
    :param inv: Whether to calculate the inversion. Default value is True
    :param out_fmt: the output format. mat: matrix format (default); row_col_val: row-column-value format;
    id_id_val: id-id-value format.
    :param npart: the number of part. If memory is not enough, please increase the value.
    :param small_val: A small vale added to the diagonal to grant the positive definite. Default value is 0.001.
    :return: return numpy array for genomic relationship matrix and its inversion. Output the matrixes into the file
    with prefix of out_file.
    """
    logging.info("{:#^80}".format("Read the SNP and calculate"))
    snp_on_disk = Bed(bed_file, count_A1=False)
    num_snp = snp_on_disk.sid_count
    num_id = snp_on_disk.iid_count
    logging.info("There are {} individuals, {} SNPs.".format(num_id, num_snp))
    snp_index = []
    for i in range(npart):
        snp_index.append(int(num_snp / npart) * i)
    snp_index.append(num_snp)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    scale = 0.0
    kin = np.zeros((num_id, num_id))
    for i in tqdm(range(npart)):
        snp_mat = snp_on_disk[:, snp_index[i]:snp_index[i + 1]].read().val
        if np.any(np.isnan(snp_mat)):
            snp_mat = impute_geno(snp_mat)
        freq = np.sum(snp_mat, axis=0) / (2 * num_id)
        freq.shape = (1, -1)
        scale_vec = 2 * freq * (1 - freq)
        scale += np.sum(scale_vec * (1 - scale_vec))
        snp_mat[snp_mat > 1.5] = 0.0
        snp_mat = snp_mat - scale_vec
        kin += np.dot(snp_mat, snp_mat.T)
    kin = kin / scale
    kin_diag = np.diag(kin)
    kin_diag = kin_diag + kin_diag * small_val
    np.fill_diagonal(kin, kin_diag)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info(
        "Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    logging.info("{:#^80}".format("Output"))
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    logging.info("The prefix of output file is " + out_file)
    fam_info.iloc[:, 1].to_csv(out_file + '.id', index=False, header=False, sep=' ')
    output_mat(kin, fam_info.iloc[:, 1], out_file + '.dgrm_as', out_fmt)
    kin_inv = None
    if inv:
        logging.info("{:#^80}".format("Calculate the inversion of kinship"))
        clock_t0 = time.perf_counter()
        cpu_t0 = time.process_time()
        kin_inv = linalg.inv(kin)
        clock_t1 = time.perf_counter()
        cpu_t1 = time.process_time()
        logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0,
                                                                                   cpu_t1 - cpu_t0))
        logging.info("{:#^80}".format("Output the inversion"))
        logging.info("The prefix of output file is: " + out_file)
        output_mat(kin_inv, fam_info.iloc[:, 1], out_file + '.dgiv_as', out_fmt)
    return kin, kin_inv



import sys
import math
import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, isspmatrix, hstack, vstack
import datetime
import logging


def _emai(y, xmat, zmat, ag_inv, init=None, maxiter=30, cc_par=1.0e-8, cc_gra=1.0e-6, em_weight_step=0.01,
          out_file='unbalance_varcom'):
    def init_var(cov_dim):
        """
        initialize the variances
        """
        y_var = np.var(y)/(cov_dim[0] + cov_dim[1] + 1)
        add_cov = np.diag([y_var]*cov_dim[0])
        add_ind = np.tril_indices_from(add_cov)
        per_cov = np.diag([y_var]*cov_dim[1])
        per_ind = np.tril_indices_from(per_cov)
        cov_ind = [1] * len(add_ind[0]) + [2] * len(per_ind[0]) + [3]
        cov_ind_i = list(add_ind[0] + 1) + list(per_ind[0] + 1) + [1]
        cov_ind_j = list(add_ind[1] + 1) + list(per_ind[1] + 1) + [1]
        var_com = list(add_cov[add_ind]) + list(per_cov[per_ind]) + [y_var]
        var_df = {'var': cov_ind,
                  "vari": cov_ind_i,
                  "varj": cov_ind_j,
                  "var_com": var_com}
        var_df = pd.DataFrame(var_df, columns=['var', "vari", "varj", "var_com"])
        return np.array(var_com), var_df
    cov_dim = [len(zmat[0]), len(zmat[1])]  # the dimension for the covariance matrix
    var_com, var_df = init_var(cov_dim)
    if init is not None:
        if len(var_com) != len(init):
            logging.error('The length of initial variances should be {}'.format(len(var_com)))
            sys.exit()
        else:
            var_com = np.array(init)
            var_df['var_com'] = var_com

    def effect_index():
        """
        return the vector of indexes for all effects
        """
        eff_ind = [[0, xmat.shape[1]]]
        for i in range(len(zmat)):
            temp = [eff_ind[i][-1]]
            for j in range(len(zmat[i])):
                temp.append(temp[-1] + zmat[i][j].shape[1])
            eff_ind.append(temp)
        return eff_ind
    eff_ind = effect_index()

    logging.info('***prepare the MME***')
    zmat_con_lst = [hstack(zmat[0]), hstack(zmat[1])]  # combined random matrix
    zmat_con = hstack(zmat_con_lst)  # design matrix for random effects
    wmat = hstack([xmat, zmat_con])  # merged design matrix
    cmat_pure = np.dot(wmat.T, wmat)  # C matrix
    rhs_pure = wmat.T.dot(y)  # right hand
    logging.info('initialize the variate')
    num_record = y.shape[0]
    iter_count = 0
    cc_par_val = 1000.0
    cc_gra_val = 1000.0
    delta = 1000.0
    var_com_update = var_com * 1000
    logging.info("initial variances: {}".format(var_df))

    def pre_cov_inv(cov_dim, var_com):
        """
        return the inversion of covariance matrix. If one covariance matrix is not positive, return None.
        """
        cov_inv = []  # the inversion of covariance matrix list
        if var_com[-1] <= 0:
            return None
        for i in range(len(cov_dim)):
            covar = np.zeros((cov_dim[i], cov_dim[i]))
            covar[np.tril_indices(cov_dim[i])] = np.array(var_df[var_df['var']==(i+1)].iloc[:, -1])
            covar = np.add(covar, np.tril(covar, -1).T)
            try:
                linalg.cholesky(covar)
            except:
                return None
            covar = linalg.inv(covar)
            cov_inv.append(covar)
        return cov_inv
    cov_inv = pre_cov_inv(cov_dim, var_com)
    # cov_inv = pre_covi_mat(cov_dim, var_com_update)
    if cov_inv is None:
        logging.error("ERROR: Initial variances is not positive define, please check!")
        sys.exit()

    def cal_cmat_inv(cmat):
        """
        return the inversion of coefficient matrix
        """
        cmat[eff_ind[1][0]:eff_ind[1][-1], eff_ind[1][0]:eff_ind[1][-1]] = \
            cmat[eff_ind[1][0]:eff_ind[1][-1], eff_ind[1][0]:eff_ind[1][-1]] + linalg.kron(cov_inv[0], ag_inv)
        cmat[eff_ind[2][0]:eff_ind[2][-1], eff_ind[2][0]:eff_ind[2][-1]] = \
            cmat[eff_ind[2][0]:eff_ind[2][-1], eff_ind[2][0]:eff_ind[2][-1]] + \
            sparse.kron(cov_inv[1], sparse.eye(ag_inv.shape[0], format="csr"))
        cmat_inv = linalg.inv(cmat)
        return cmat_inv

    def cal_fd_mat(cmat):
        """
        Calculate the first-order derivative
        """
        fd_mat = []
        # addtive part
        qnum = zmat_con_lst[0].shape[1] / cov_dim[0]
        tmat = np.zeros((cov_dim[0], cov_dim[0]))
        eff_lst = []
        for j in range(cov_dim[0]):
            for k in range(j + 1):
                tmat[j, k] = np.sum(np.multiply(ag_inv, cmat[eff_ind[1][j]:eff_ind[1][j + 1],
                                                        eff_ind[1][k]:eff_ind[1][k + 1]]))
                tmat[k, j] = tmat[j, k]
            eff_lst.append(eff[eff_ind[1][j]:eff_ind[1][j + 1], :])
        eff_mat = np.dot(np.concatenate(eff_lst, axis=1), cov_inv[0])
        fd_val = qnum * cov_inv[0] - np.dot(np.dot(cov_inv[0], tmat), cov_inv[0]) - np.dot(np.dot(eff_mat.T, ag_inv), eff_mat)
        fd_val[np.tril_indices(cov_dim[0], k=-1)] = 2.0 * fd_val[np.tril_indices(cov_dim[0], k=-1)]
        fd_mat.extend(list(-0.5 * fd_val[np.tril_indices(cov_dim[0], k=0)]))
        # per part
        eff_lst = []
        per_kin = sparse.eye(ag_inv.shape[0], format="csr")
        for j in range(cov_dim[1]):
            for k in range(j + 1):
                tmat[j, k] = np.sum(per_kin.multiply(cmat[eff_ind[2][j]:eff_ind[2][j + 1],
                                                    eff_ind[2][k]:eff_ind[2][k + 1]]))
                tmat[k, j] = tmat[j, k]
            eff_lst.append(eff[eff_ind[2][j]:eff_ind[2][j + 1], :])
        eff_mat = np.dot(np.concatenate(eff_lst, axis=1), cov_inv[1])
        fd_val = qnum * cov_inv[1] - np.dot(np.dot(cov_inv[1], tmat), cov_inv[1]) \
               - np.dot(eff_mat.T, per_kin.dot(eff_mat))
        fd_val[np.tril_indices(cov_dim[1], k=-1)] = 2.0 * fd_val[np.tril_indices(cov_dim[1], k=-1)]
        fd_mat.extend(list(-0.5 * fd_val[np.tril_indices(cov_dim[1], k=0)]))
        # error part
        fd_val = -0.5 * (num_record / var_com[-1] - np.sum((wmat.T.dot(wmat)).multiply(cmat)) \
                       / (var_com[-1] * var_com[-1]) - np.sum(np.dot(e.T, e) / (var_com[-1] * var_com[-1])))
        fd_mat.append(fd_val)
        return np.array(fd_mat)

    def cal_ai_mat():
        """
        Calculate the AI matrix.
        """
        wv = []
        for i in range(len(cov_dim)):
            dial = sparse.eye(zmat_con_lst[i].shape[1] / cov_dim[i], dtype=np.float64)
            for j in range(cov_dim[i]):
                for k in range(j + 1):
                    var_fd = np.zeros((cov_dim[i], cov_dim[i]))  # Notice
                    var_fd[j, k] = var_fd[k, j] = 1.0
                    temp = sparse.kron(np.dot(var_fd, cov_inv[i]), dial)
                    temp = temp.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :])
                    temp = zmat_con_lst[i].dot(temp)
                    wv.append(temp)
        wv.append(e / var_com[-1])
        qmat = np.concatenate(wv, axis=1)
        qrq = np.divide(np.dot(qmat.T, qmat), var_com[-1])
        left = np.divide(wmat.T.dot(qmat), var_com[-1])
        eff_qmat = np.dot(cmat, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_qmat))
        ai_mat = 0.5 * ai_mat
        return ai_mat

    def cal_em_mat():
        """
        Calculate the em matrix.
        """
        num_var_com = len(var_com)
        em_mat = np.zeros((num_var_com, num_var_com))
        b = 0
        cov_mat = []
        for i in range(len(cov_dim)):
            ind = np.tril_indices(cov_dim[i])
            temp = np.zeros((len(ind[0]), len(ind[0])))
            q = zmat_con_lst[i].shape[1] / cov_dim[i]
            a = b
            b = b + len(ind[0])
            covar = np.zeros((cov_dim[i], cov_dim[i]))
            covar[ind] = var_com[a:b]
            covar = np.add(covar, np.tril(covar, -1).T)
            cov_mat.append(covar)
            for j in range(len(ind[0])):
                for k in range(j + 1):
                    temp[j, k] = (cov_mat[i][ind[0][j], ind[0][k]] * cov_mat[i][ind[1][j], ind[1][k]] + \
                                  cov_mat[i][ind[0][j], ind[1][k]] * cov_mat[i][ind[1][j], ind[0][k]]) / (2.0 * q)
            em_mat[a:b, a:b] = temp.copy()
        em_mat[-1, -1] = (var_com[-1] * var_com[-1]) / num_record
        em_mat = 2.0 * em_mat
        em_mat += np.tril(em_mat, k=-1).T
        return linalg.inv(em_mat)

    while iter_count < maxiter:
        iter_count += 1
        logging.info('***Start the iteration: ' + str(iter_count) + ' ***')
        logging.info("Prepare the coefficient matrix")
        cmat = (cmat_pure.multiply(1.0 / var_com[-1])).toarray()
        cmat = cal_cmat_inv(cmat)
        rhs_mat = np.divide(rhs_pure, var_com[-1])
        eff = np.dot(cmat, rhs_mat)
        e = y - np.dot(xmat, eff[:eff_ind[0][1], :]) - zmat_con.dot(
            eff[eff_ind[0][1]:, :])
        logging.info("first-order derivative")
        fd_mat = cal_fd_mat(cmat)
        np.savetxt(out_file + '.fd', fd_mat)
        logging.info("AI matrix")
        ai_mat = cal_ai_mat()
        np.savetxt(out_file + '.ai', ai_mat)
        logging.info("EM matrix")
        em_mat = cal_em_mat()
        np.savetxt(out_file + '.em', em_mat)
        # Increase em weight to guarantee variances positive
        gamma = -em_weight_step
        while gamma < 1.0:
            gamma = gamma + em_weight_step
            if gamma >= 1.0:
                gamma = 1.0
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_update = var_com + delta
            var_df['var_com'] = var_com_update
            cov_inv = pre_cov_inv(cov_dim, var_com_update)
            if cov_inv is not None:
                logging.info('EM weight value: ' + str(gamma))
                break
        var_df.to_csv(out_file + '.var', sep='\t', index=False)
        if cov_inv is None:
            logging.error("ERROR: Updated variances is not positive define!")
            sys.exit()
        # Convergence criteria
        cc_par_val = np.sum(pow(delta, 2)) / np.sum(pow(var_com_update, 2))
        cc_par_val = np.sqrt(cc_par_val)
        cc_gra_val = np.sqrt(np.sum(pow(fd_mat, 2))) / len(var_com)
        var_com = var_com_update.copy()
        logging.info("Change in parameters: {}, Norm of gradient vector: {}".format(cc_par_val, cc_gra_val))
        if cc_par_val < cc_par and cc_gra_val < cc_gra:
            break
    if cc_par_val < cc_par and cc_gra_val < cc_gra:
        logging.info("Variances Converged")
    else:
        logging.info("Variances not Converged")
    var_df['var_com'] = var_com
    return var_df
